@@ -38,13 +38,30 @@
  * file access
  */
 
-#define file_close(X) X->close(X)
+#define file_close(X)    X->close(X)
 #define file_seek(X,Y,Z) X->seek(X,Y,Z)
-#define file_tell(X) X->tell(X)
-#define file_read(X,Y,Z) X->read(X,Y,Z)
+#define file_tell(X)     X->tell(X)
 
+static inline int64_t file_read(AACS_FILE_H *fp, void *buf, int64_t size) {
+    return fp->read(fp, buf, size);
+}
+
+static inline int64_t file_write(AACS_FILE_H *fp, const void *buf, int64_t size) {
+    return fp->write ? fp->write(fp, buf, size) : 0;
+}
+
+
+BD_PRIVATE int64_t file_size(AACS_FILE_H *fp);
 
 BD_PRIVATE extern AACS_FILE_H *(*file_open)(const char* filename, const char *mode);
 
+/*
+ * local filesystem
+ */
+
+BD_PRIVATE int file_unlink(const char *file);
+BD_PRIVATE int file_path_exists(const char *path);
+BD_PRIVATE int file_mkdir(const char *dir);
+BD_PRIVATE int file_mkdirs(const char *path);
 
 #endif /* FILE_H_ */
