@@ -23,6 +23,7 @@
 #include "util/attributes.h"
 
 #include <stdint.h>
+#include <stddef.h> /* size_t */
 
 /* struct holding a digit and key pair for <ENTRY NUMBER> - <ENTRY> entries */
 typedef struct digit_key_pair_t digit_key_pair;
@@ -41,6 +42,7 @@ struct digit_key_pair_list_t
 };
 
 /* date entry struct */
+#if 0
 typedef struct date_entry_t date_entry;
 struct date_entry_t
 {
@@ -48,6 +50,7 @@ struct date_entry_t
   unsigned int month;
   unsigned int day;
 };
+#endif
 
 /* dk entry */
 typedef struct dk_entry dk_list;
@@ -84,14 +87,14 @@ typedef struct title_entry_t title_entry;
 struct title_entry_t
 {
   uint8_t discid[20];
-  char *title;
-  date_entry date;
+  //char *title;
+  //date_entry date;
   char *mek;
   char *vid;
-  digit_key_pair_list *bn;
-  char *vuk;
-  digit_key_pair_list *pak;
-  digit_key_pair_list *tk;
+  //digit_key_pair_list *bn;
+  uint8_t vuk[16];
+  //digit_key_pair_list *pak;
+  //digit_key_pair_list *tk;
   digit_key_pair_list *uk;
 };
 
@@ -127,12 +130,10 @@ BD_PRIVATE int   keycache_save(const char *type, const uint8_t *disc_id,
 BD_PRIVATE int   keycache_find(const char *type, const uint8_t *disc_id,
                                  uint8_t *key, unsigned int len);
 
-BD_PRIVATE int cache_get(const char *name, uint32_t *version, uint32_t *len, void *buf); /* use buf=NULL to get version and size */
+BD_PRIVATE int cache_get(const char *name, uint32_t *version, uint32_t *len, void *buf, size_t buf_size); /* use buf=NULL to get version and size */
 BD_PRIVATE int cache_save(const char *name, uint32_t version, const void *data, uint32_t len);
 BD_PRIVATE int cache_remove(const char *name);
 
-BD_PRIVATE void *cache_get_or_update(const char *type, const void *data, uint32_t *len, uint32_t version);
-
 BD_PRIVATE int config_get(const char *name, uint32_t *len, void *buf); /* use buf=NULL to get size */
 BD_PRIVATE int config_save(const char *name, const void *data, uint32_t len);
 
