@@ -20,12 +20,12 @@
 #ifndef AACS_PATH_H
 #define AACS_PATH_H
 
-#include <util/attributes.h>
+#include "util/attributes.h"
 
 #ifndef AACS_PATH_MAX
 #define AACS_PATH_MAX 1024
 #endif
 
-AACS_PRIVATE char *aacs_resolve_path(const char *path, char *resolved_path);
+BD_PRIVATE char *aacs_resolve_path(const char *path, char *resolved_path);
 
 #endif /* AACS_PATH_H */
